gstrGameVersion = "0.07"

inspect = require 'inspect'
-- https://github.com/kikito/inspect.lua

TLfres = require "tlfres"
-- https://love2d.org/wiki/TLfres
-- screen stack

anim8 = require 'anim8'
-- https://github.com/kikito/anim8

bitser = require 'bitser'

nativefs = require("nativefs")
-- https://github.com/megagrump/nativefs

lovelyToasts = require("lovelyToasts")
-- https://github.com/Loucee/Lovely-Toasts  


cobjs = require "createobjects"
dobjs = require "drawobjects"
fun = require "randomfunctions"

cf = require "commonfunctions"
--ft = require "foxtree"
--bt = require "behaviortree"
--enum = require "enum"

garrCurrentScreen = {}
garrGameSettings = {}		-- global game settings
garrTank = {}	
garrFish = {}
garrFood = {}
garrFilter = {}
garrWave = {}		-- intentionally global for persistence	
	

gbolDebug = true
gbolPaused = false
gbolShowData = false

gstrCurrentScreen = ""

gconstGlobalTimer = 3600

gfltGlobalTimer = gconstGlobalTimer			-- 1 hour
gfltMovementTimerFish = 0.5
gfltMovementWave = 0.75
gintSineOffset = 0
gintHours = 0								-- number of hours sim has been running

gintSoundVolume =  1

function GetTapWater(tank, volume)
-- gets tap water into a temporary storage
-- ph = 7.8
-- nitrate = 0.5535 mg/L
-- nitrite = 0.295 mg/L
-- chlorine = 4.1 mg/L
-- hardness = 110 mg/L


	local intBucketSize = 10	-- litres
	
	if garrTank[tank].watervolume <= intBucketSize - volume then
		garrTank[tank].watervolume = garrTank[tank].watervolume + volume
		
		garrTank[tank].GH = garrTank[tank].GH + 110
		garrTank[tank].chlorine = garrTank[tank].chlorine + 2.1
		garrTank[tank].chloramine = garrTank[tank].chloramine + 0.002
		garrTank[tank].nitrite = garrTank[tank].nitrite + 0.295
		garrTank[tank].nitrate = garrTank[tank].nitrate + 0.5535
		
		-- need to get the average PH
		local phdelta = 7.8 - garrTank[tank].PH		-- get the difference in PH

		if (garrTank[tank].watervolume - volume) > 0 then
			local phdelta = phdelta / (garrTank[tank].watervolume - volume)	-- adjust that delta by the volume already in the bucket
			garrTank[tank].PH = garrTank[tank].PH + phdelta
		else
			garrTank[tank].PH = 7.8
		end
	end
end

function RemoveTapWater(tank, volume)

	local origvolume = garrTank[tank].watervolume
	local percentremoved = volume/origvolume		-- eg if 1L is removed from 6L then volume affected = 1/6 = 16%
	
	if garrTank[tank].watervolume >= volume then
		garrTank[tank].watervolume = garrTank[tank].watervolume - volume

		garrTank[tank].GH = garrTank[tank].GH - (garrTank[tank].GH * percentremoved )
		garrTank[tank].chlorine = garrTank[tank].chlorine - (garrTank[tank].chlorine * percentremoved )
		garrTank[tank].chloramine = garrTank[tank].chloramine - (garrTank[tank].chloramine * percentremoved )
		garrTank[tank].nitrite = garrTank[tank].nitrite - (garrTank[tank].nitrite * percentremoved )
		garrTank[tank].nitrate = garrTank[tank].nitrate - (garrTank[tank].nitrate * percentremoved )
		
	end
	
	if garrTank[tank].watervolume <= 0 then
		garrTank[tank].watervolume = 0
		garrTank[tank].GH = 0
		garrTank[tank].PH = 0
		garrTank[tank].chlorine = 0
		garrTank[tank].chloramine = 0
		garrTank[tank].nitrite = 0
		garrTank[tank].nitrate = 0	
	end

end

function AgeWater(tank, duration)
-- age the water by duration (hours) so that chlorine evaporates

	garrTank[tank].chlorine = garrTank[tank].chlorine * (21/24)


end

function AddWater(maintank, bucket, volume)
-- adds the required volume of water from temp storage into the tank
-- recalculates chemistry

	-- check there is room to transfer water
	if garrTank[bucket].watervolume >= volume then
		if garrTank[maintank].watervolume <= garrTank[maintank].totalvolume - volume then
		
			local origmaintankvolume = garrTank[maintank].watervolume
			local origvolume = garrTank[bucket].watervolume
			local percentremoved = volume/origvolume		-- eg if 1L is removed from 6L then volume affected = 1/6 = 16%		
		
			garrTank[bucket].watervolume = garrTank[bucket].watervolume - volume
			garrTank[maintank].watervolume = garrTank[maintank].watervolume + volume
			
			local ghmoved = (garrTank[bucket].GH * percentremoved )
			local chlorinemoved = (garrTank[bucket].chlorine * percentremoved )
			local chloraminemoved = (garrTank[bucket].chloramine * percentremoved )
			local nitritemoved = (garrTank[bucket].nitrite * percentremoved )
			local nitratemoved = (garrTank[bucket].nitrate * percentremoved )
			
			
			-- garrTank[bucket].GH = garrTank[bucket].GH - ghmoved		--! not right
			garrTank[bucket].chlorine = garrTank[bucket].chlorine - chlorinemoved
			garrTank[bucket].chloramine = garrTank[bucket].chloramine - chloraminemoved
			garrTank[bucket].nitrite = garrTank[bucket].nitrite - nitritemoved
			garrTank[bucket].nitrate = garrTank[bucket].nitrate - nitratemoved	

			-- garrTank[maintank].GH = garrTank[maintank].GH + ghmoved
			garrTank[maintank].chlorine = garrTank[maintank].chlorine + chlorinemoved
			garrTank[maintank].chloramine = garrTank[maintank].chloramine + chloraminemoved
			garrTank[maintank].nitrite = garrTank[maintank].nitrite + nitritemoved
			garrTank[maintank].nitrate = garrTank[maintank].nitrate + nitratemoved	
			
			if origmaintankvolume <= 0 then
				garrTank[maintank].PH = garrTank[bucket].PH
			else
				-- adjust by proportion
				local phdelta = garrTank[maintank].PH - garrTank[bucket].PH		-- get the difference in PH
				local portion = volume/origvolume
				phdelta = phdelta * portion
				garrTank[maintank].PH = garrTank[maintank].PH - phdelta
			end
		end
	end	
end

function RemoveWater(tank, volume)
-- removes the required volume of water from the tank
-- recalculates chemistry

	if volume > garrTank[tank].watervolume then volume = garrTank[tank].watervolume end
	
	local percentremoved = volume/garrTank[tank].watervolume
	
	garrTank[tank].watervolume = garrTank[tank].watervolume - volume
	garrTank[tank].chlorine = garrTank[tank].chlorine * (1 - percentremoved )
	garrTank[tank].chloramine = garrTank[tank].chloramine * (1 - percentremoved )
	garrTank[tank].nitrite = garrTank[tank].nitrite * (1 - percentremoved )
	garrTank[tank].nitrate = garrTank[tank].nitrate * (1 - percentremoved )
	garrTank[tank].ammonia = garrTank[tank].ammonia * (1 - percentremoved )
	
	-- if fish is above the waterlevel then drop them down
	for k,v in ipairs(garrFish) do
		if v.y < fun.GetWaterY() then v.y = fun.GetWaterY() end
	end
	
	
	-- if all water gone then kill fish
	if garrTank[tank].watervolume <= 0 then
		garrTank[tank].watervolume = 0

		garrTank[tank].GH = 0
		garrTank[tank].PH = 0
		garrTank[tank].chlorine = 0
		garrTank[tank].chloramine = 0
		garrTank[tank].nitrite = 0
		garrTank[tank].nitrate = 0
		garrTank[tank].ammonia = 0		
		garrFish = {}
	end
end

local function GetValidX()
	return (TankX1 + love.math.random(50,gintdrawingwidth - 50))
end

local function GetValidY(myfish)
	
	local litreheight = gintdrawingheight / garrTank[1].totalvolume
	local wheight = litreheight * garrTank[1].watervolume		-- water height eg 2L = 60. Gives pixels	
	return (love.math.random((TankY1 + gintdrawingheight - wheight) , (TankY1 + gintdrawingheight))) - (128 * myfish.growth) -- 128 is the sprite height
end

function AddFish(species)
-- adds the required species to the tank just once
	if garrTank[1].watervolume > 0 then
		cobjs.CreateFish()
		local lastfish = #garrFish
		garrFish[lastfish].x = GetValidX()
		garrFish[lastfish].y = GetValidY(garrFish[lastfish])

	end
end

function GrowBacteriaA()
-- grow the colony of bacteria A
-- growth requires ammonia

	-- convert ammonia to nitrite
	local adjustedbacteriaA = garrFilter[1].bacteriaA / 1000
	local adjustment = math.min(adjustedbacteriaA,garrTank[1].ammonia )

	garrTank[1].ammonia = garrTank[1].ammonia - adjustment
	garrTank[1].nitrite = garrTank[1].nitrite + adjustment
	garrFilter[1].bacteriaA = garrFilter[1].bacteriaA + adjustment * 1000
	
	-- convert nitrite into nitrate
	local adjustedbacteriaB = garrFilter[1].bacteriaB / 1000
	local adjustment = math.min(adjustedbacteriaB,garrTank[1].nitrite )	
	
	garrTank[1].nitrite = garrTank[1].nitrite - adjustment
	garrTank[1].nitrate = garrTank[1].nitrate + adjustment
	garrFilter[1].bacteriaB = garrFilter[1].bacteriaB + adjustment * 1000

end

function GrowBacteriaAB()
-- grow the colony of bacteria B
-- growth requires nitrite

end

function IncreaseAmmonia()
-- increases ammonia based on fish and excess food

end

function ConvertToNitrite()
-- decrease ammonia and increase nitrite
-- requries bacteria A


end

function ConvertToNitrate()
-- decrease nitrite and increase nitrate
-- requries bacteria B


end

function MoveFish()

	for k,v in ipairs(garrFish) do
	
		if v.dead then
			if v.y > fun.GetWaterY() then
				v.y = v.y -1
			end

		else
		
			if v.hunting == false and v.fullness < 100 and #garrFood > 0 and v.targetx == 0 then
				-- hunt food
				local rndnum = love.math.random(1, #garrFood)
				v.targetx = garrFood[rndnum].x
				v.targety = GetValidY(v)
				v.hunting = true
			elseif v.targetx ~= 0 then
				-- have target, do nothign
			else
				-- random location
				v.targetx = GetValidX()
				v.targety = GetValidY(v)
				v.hunting = false
			end

			-- determine facing
			if v.targetx > v.x and v.heading == "left" then
				v.heading = "right"
			end
			if v.targetx < v.x and v.heading == "right" then
				v.heading = "left"
			end			

		
			local speed = 1
			if v.hunting then speed = 2 end
			local chlorineppm = cf.round((garrTank[1].chlorine / garrTank[1].watervolume),3)
			local chloramineppm = cf.round((garrTank[1].chloramine / garrTank[1].watervolume),3)
			if chlorineppm > 0.1 or chloramineppm > 0.0035 then speed = 0.5 end

			if v.targetx > v.x then
				v.x = v.x + speed
			end
			if v.targetx < v.x then
				v.x = v.x - speed
			end
			if v.targety > v.y then
				v.y = v.y + speed
			end
			if v.targety < v.y then
				v.y = v.y - speed
			end
			
			if chlorineppm > 0.1 or chloramineppm > 0.0035 then
				-- gasping for air
				-- move to top
				v.y = v.y - speed
			end			
			

			-- check for arrival
			if math.abs(v.targetx - v.x) <= 10 and math.abs(v.targety - v.y) < 10 then
				-- arrived
				v.targetx = 0
				v.targety = 0
				v.hunting = false
			end
			
			-- see if food nearby
			if v.hunting then
				for i,j in ipairs(garrFood) do
					-- scan all food and see if close
					local dist = cf.GetDistance(v.x, v.y, j.x, j.y)
					if dist <= 10 then
						-- yum
						table.remove(garrFood, i)
						v.fullness = v.fullness + 20
						v.hunting = false
						garrTank[1].ammonia = garrTank[1].ammonia + 1
					end
				end
			end		
	
		end

	end

end

function MoveFood()
-- make food fall

	for k,v in ipairs(garrFood) do
		v.y = v.y + 1
		if v.y > fun.GetTankBottomY() then
			table.remove(garrFood,k)
			garrTank[1].ammonia = garrTank[1].ammonia + 1
		end
	end
end

function GrowFish()

	for k,v in ipairs(garrFish) do
		if v.fullness > 50 then
			v.growth = v.growth + 0.01
			if v.growth > 0.30 then v.growth = 0.30 end
		end
	end
end

function HungryFish()
	for k,v in ipairs(garrFish) do
		if v.fullness > 0 then
			v.fullness = v.fullness - 4	-- lose 4.2% fullness per hour
			if v.fullness < 0 then v.fullness = 0 end
		end
	end
end

function KillFish()

	for k,v in ipairs(garrFish) do
		-- hunger
		if v.fullness == 0 then
			v.dead = true
		end
		
		-- ammonia
		local ammoniappm = cf.round((garrTank[1].ammonia / garrTank[1].watervolume),3)
		if ammoniappm > 1 then
			-- random chance of dying
			local diechance = math.floor(ammoniappm * 20)
			if love.math.random(0,100) > diechance then
				v.dead = true
			end
		end
		
		-- chlorine
		local chlorineppm = cf.round((garrTank[1].chlorine / garrTank[1].watervolume),3)
		if chlorineppm > 0.3 then
			v.dead = true
		end
		
		-- chloramine
		local chloramineppm = cf.round((garrTank[1].chloramine / garrTank[1].watervolume),3)
		if chloramineppm > 0.0055 then
			v.dead = true
		end		
	end
end

function Dechlorinate()
	local tank
	
	if garrCurrentScreen[#garrCurrentScreen] == "Tank" then
		tank = 1
	end
	if garrCurrentScreen[#garrCurrentScreen] == "Bucket" then
		tank = 2
	end
	
	assert(tank ~= nil)
	if garrTank[tank].watervolume > 0 then
		garrTank[tank].chloramine = 0
		garrTank[tank].chlorine = 0
		-- garrTank[tank].ammonia = 0
		lovelyToasts.show("Dechlorinator added to water", 3, "bottom", gintScreenWidth)		-- text, duration, y or "bottom", screenwidth
	else
		lovelyToasts.show("Add water first", 3, "bottom", gintScreenWidth)		-- text, duration, y or "bottom", screenwidth
	end
end

function AddFilter()

	if #garrFilter == 0 then
		cobjs.CreateFilter()
	end

end

function EvaporateWater()

	garrTank[1].watervolume = garrTank[1].watervolume - (1 / 168)  -- this is 1 litre per week
	if garrTank[1].watervolume <= 0 then garrTank[1].watervolume = 0 end

end

function GenerateWaves()

	local offset = 50		-- for that sweet 3D look
	local filterwidth = 50
	local litreheight = gintdrawingheight / garrTank[1].totalvolume
	local wheight = litreheight * garrTank[1].watervolume		-- water height eg 2L = 60. Gives pixels
	
	local x2 = (TankX1 + offset + filterwidth ) * 1.33
	local y2 = TankY1 + gintdrawingheight - wheight
	
	for i = x2 , (x2 * 2) do
	
		local rndstart = love.math.random(0, offset)
		local rndlength = love.math.random(1, 3)

		x4 = i + rndstart
		y4 = y2 - rndstart
		
		x5 = x4 + rndlength
		y5 = y4 - rndlength
		
		if y5 < y4 - offset then y5 = y4 - offset end
	
		garrWave[i] = {}
		garrWave[i].x4 = x4
		garrWave[i].y4 = y4
		garrWave[i].x5 = x5
		garrWave[i].y5 = y5
		
	end
end

function ExecuteOneHour()

		AgeWater(2, 1)		-- tank/hours
		AgeWater(1, 1)		-- tank/hours
		EvaporateWater()
		GrowFish()
		HungryFish()
		KillFish()
		--!KillBacteria()
		if #garrFilter > 0 then
			GrowBacteriaA()
		end
		
		gintHours = gintHours + 1
		

end

function love.keyreleased( key, scancode )
	if key == "escape" then
		fun.RemoveScreen()
	end
	
	if key == "lalt" or key == "ralt" then
		gbolShowData = not gbolShowData
	end
	
	if key == "1" then
		fun.AddScreen("Bucket")
	end
	if key == "2" then
		GetTapWater(2,1)		-- add water to bucket
	end
	if key == "3" then
		RemoveTapWater(2,1)		-- remove water from bucket
	end	
	if key == "4" then
		ExecuteOneHour()
	end		
	if key == "5" then
		AddWater(1,2,1)			-- transfers 1L from tank 2 to tank 1
	end	
	if key == "6" then
		AddFish(nil)			
	end		
	if key == "7" then
		RemoveWater(1, 10)
	end	
	if key == "8" then
		if garrCurrentScreen[#garrCurrentScreen] == "Tank" then
			fun.AddScreen("AmmoniaTest")
		elseif garrCurrentScreen[#garrCurrentScreen] == "AmmoniaTest" then
			fun.RemoveScreen()
		end
	end	

	if key == "s" then
		garrGameSettings.globaltimer = gfltGlobalTimer
		garrGameSettings.hours = gintHours
		garrGameSettings.gameversion = tonumber(gstrGameVersion)

		fun.SaveGame()
		lovelyToasts.show("Game saved", 3, "bottom", gintScreenWidth)		-- text, duration, y or "bottom", screenwidth
	end		
	if key == "l" then
		fun.LoadGame()
		
		gfltGlobalTimer = garrGameSettings.globaltimer
		gintHours = garrGameSettings.hours
		
	end
	if key == "d" then
		Dechlorinate()
	end	
	if key == "f" then
		AddFilter()
	end	
	
	if key == "kp+" then
		gintSoundVolume = gintSoundVolume + 0.1
		
		garrSoundEffects[1]:stop()
		garrSoundEffects[1]:setVolume(gintSoundVolume)
		garrSoundEffects[1]:play()

	end
	
	if key == "kp-" then
		gintSoundVolume = gintSoundVolume - 0.1
		if gintSoundVolume < 0 then gintSoundVolume = 0 end
		
		garrSoundEffects[1]:stop()
		garrSoundEffects[1]:setVolume(gintSoundVolume)
		garrSoundEffects[1]:play()		

	end

end

function love.mousepressed( x, y, button, istouch, presses )

	local mousex,mousey = TLfres.getMousePosition(gintScreenWidth, gintScreenHeight)    -- lets you pretend screen is 1920 * 1080

	if button == 1 then
		-- drop fish food
		if mousex > TankX1 and mousex < (TankX1 + gintdrawingwidth) then
		
			local litreheight = gintdrawingheight / garrTank[1].totalvolume
			local wheight = litreheight * garrTank[1].watervolume		-- water height eg 2L = 60. Gives pixels
			y3 = TankY1 + gintdrawingheight - wheight
			if mousey < y3 then
				cobjs.CreateFood(mousex,mousey)
				-- clear targets for hungry fish so they can acquire food
				for k,v in ipairs(garrFish) do
					if v.hunting == false and v.fullness < 100 then
						v.targetx = 0
						v.targety = 0
					end
				end
			end
		end
	end
end

function love.load()
    if love.filesystem.isFused( ) then
        void = love.window.setMode(gintScreenWidth, gintScreenHeight,{fullscreen=false,display=1,resizable=true, borderless=false})	-- display = monitor number (1 or 2)
        gbolDebug = false
    else
        void = love.window.setMode(gintScreenWidth, gintScreenHeight,{fullscreen=false,display=2,resizable=true, borderless=false})	-- display = monitor number (1 or 2)
    end
	
	love.window.setTitle("My fish tank water simulator " .. gstrGameVersion)
	
	cobjs.CreateGameSettings()
	cobjs.CreateTank()		-- main tank
	cobjs.CreateTank()		-- bucket
	

	
	garrTank[1].totalvolume = 132
	garrTank[2].totalvolume = 10
	
	garrTank[1].watervolume = 0
	if love.filesystem.isFused( ) then
		garrTank[1].watervolume = 0
	else
		--garrTank[1].watervolume = 100
	end
	
	-- these are global values
	-- FishSpriteSheet = love.graphics.newImage("assets/fishsprites.png")
	-- FishSprite1 = cf.fromImageToQuads(FishSpriteSheet,16,32)	-- sprite width/height

    FishSpriteSheet = love.graphics.newImage("assets/fishsprites.png")
    Fishanim8 = anim8.newGrid(128, 128, FishSpriteSheet:getWidth(), FishSpriteSheet:getHeight())     -- frame width, frame height
    FishAnimationRight = anim8.newAnimation(Fishanim8('1-3',1), 1)	-- FishAnimationRight:draw(sheet, drawingx, drawingy,0,imgScale,imgScale)
																	-- FishAnimationRight:update(dt)
	FishAnimationLeft = anim8.newAnimation(Fishanim8('1-3',1), 1)
	FishAnimationLeft:flipH()
	
	FishAnimationDead = anim8.newAnimation(Fishanim8('1-3',1), 1)
	FishAnimationDead:flipV()
	FishAnimationDead:pauseAtEnd()
	
	garrSoundEffects = {}
	garrSoundEffects[1] = love.audio.newSource("assets/107126__raytarara__waterrunning.wav", "stream")
	garrSoundEffects[1]:setVolume(gintSoundVolume)
	
	
	fun.AddScreen("Tank")
	
	lovelyToasts.options.queueEnabled = true

end

function love.draw(dt)

	TLfres.beginRendering(gintScreenWidth,gintScreenHeight)

	if garrCurrentScreen[#garrCurrentScreen] == "Tank" then
		dobjs.drawtank()
		if #garrFilter > 0 then
			dobjs.drawfilter()
			if garrTank[1].watervolume > 0 then
				dobjs.drawwave()
			end
		end
		if garrTank[1].watervolume > 0 then
			dobjs.drawwater()
			
		end
		dobjs.drawinstructions()		-- this function will know what screen we are on
		dobjs.drawfish()
		dobjs.drawfood()
		
		
		
		
	end
	if garrCurrentScreen[#garrCurrentScreen] == "Bucket" then
		dobjs.drawbucket()
		dobjs.drawinstructions()		-- this function will know what screen we are on
	end
	
	if garrCurrentScreen[#garrCurrentScreen] == "AmmoniaTest" then
		dobjs.drawammoniastrip(1)
	end
	
	lovelyToasts.draw()
	
	TLfres.endRendering({0, 0, 0, 1})
end

function love.update(dt)

	gstrCurrentScreen = garrCurrentScreen[#garrCurrentScreen]
	
	gfltGlobalTimer = gfltGlobalTimer - dt
	if gfltGlobalTimer <= 0 then
		ExecuteOneHour()
		gfltGlobalTimer = gconstGlobalTimer
	end
	
	gfltMovementTimerFish = gfltMovementTimerFish - dt
	if gfltMovementTimerFish <= 0 then
		gfltMovementTimerFish = 0.25
		MoveFish()
		MoveFood()
	end
	
	gfltMovementWave = gfltMovementWave - dt
	if gfltMovementWave <= 0 then 
		gfltMovementWave = 0.75
		gintSineOffset = gintSineOffset + 1

		garrWave = {}
		
		GenerateWaves()
		
		if gintSineOffset > 5 then 
			gintSineOffset = 1 
			
		end
	end
	
	local intCount = love.audio.getActiveSourceCount()
	if intCount == 0 and #garrFilter > 0 then

		garrSoundEffects[1]:play()
	end
	
	
	FishAnimationRight:update(dt)
	FishAnimationLeft:update(dt)
	lovelyToasts.update(dt)
	

end



-- fish tank = 36 * 15 * 15 inches
-- 915mm * 380mm * 380mm
-- volume = 8100 qubic inches
-- volume = 132126 qubic cm
-- volume 3D = 132 litres
-- volume 2D = 3
-- area = 34,770 square cm
-- area = 347.7 square metre
-- 











